/////////////////////////////////////////////////////////////////////////////
///
/// @file opensqr.h
///
/// An open-interval version of the square function.
/// This file contains the definition of the square function f (x) = x^2
/// in the interval arithmetic based on open intervals,
/// as opposed to closed intervals.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2007-2010 by Pawel Pilarczyk.
//
// This file is part of my research program package.  This is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on March 19, 2009. Last revision: March 19, 2009.


#ifndef _FINRESDYN_OPENSQR_H_
#define _FINRESDYN_OPENSQR_H_


// include local header files
#include "rounding.h"


// --------------------------------------------------
// ------ a function that computes the square -------
// --------------------------------------------------

/// Computes the square f (x) = x^2 in the open interval arithmetic.
template <class NumType>
void openSqr (const NumType &xLeft, const NumType &xRight,
	NumType &yLeft, NumType &yRight)
{
	// extract the rounding class to shorten the notation
	typedef tRounding<NumType> rnd;

	// make sure the input interval is correct
	if (xLeft >= xRight)
		throw "Wrong input interval for openSqr.";

	if (xLeft >= 0)
	{
		yLeft = rnd::mul_down (xLeft, xLeft);
		yRight = rnd::mul_up (xRight, xRight);
	}
	else if (xRight <= 0)
	{
		yLeft = rnd::mul_down (xRight, xRight);
		yRight = rnd::mul_up (xLeft, xLeft);
	}
	else
	{
		yLeft = -rnd::min_number ();
		yRight = (-xLeft < xRight) ? rnd::mul_up (xRight, xRight) :
			rnd::mul_up (xLeft, xLeft);
	}
	return;
} /* openSqr */


#endif // _FINRESDYN_OPENSQR_H_

